#include <LiquidCrystal_I2C.h>
#include "DHT.h"
#include <TimerOne.h>

#define DHTTYPE DHT11
#define DHTPIN 2
#define LIGHTPIN 8
#define SOILHUMIDITYPIN A0
#define LEDPIN 13
#define MEASURE_PERIOD 600000 // 10 minutes
#define NO_LIGHT_PERIOD 3600000 // 1 h

LiquidCrystal_I2C lcd(0x27,20,4);
DHT dht(DHTPIN, DHTTYPE);

// State of parameters
bool good_temp, good_light, good_soil, plant_mood;
volatile bool measure_flag = false; // Flag for measurement timing
volatile int dark_start_time = 0;

void init_lcd() {
  lcd.init();
  lcd.backlight();
  lcd.setCursor(0,0);
}

void configure_pins() {
  pinMode(LIGHTPIN, INPUT);
  pinMode(SOILHUMIDITYPIN, INPUT);
  pinMode(LEDPIN, OUTPUT);
}

void setup() {
  Serial.begin(9600);

  // Configure LCD related
  init_lcd();
  dht.begin();

  // Sets plant sensors pins on input mode
  configure_pins();

  // Set up Timer1 for periodic measurements
  Timer1.initialize(MEASURE_PERIOD); // Initialize Timer1 to overflow every MEASURE_PERIOD
  Timer1.attachInterrupt(timerIsr); // Attach interrupt service routine

  // Set up external interrupt for light sensor
  attachInterrupt(digitalPinToInterrupt(LIGHTPIN), lightChangeIsr, CHANGE);
}

void print_details_on_lcd(bool good_temp, bool good_light, bool good_soil) {
  lcd.setCursor(0,0);

  lcd.print("temp->");
  lcd.print(good_temp ? "P" : "F");

  lcd.print(" light->");
  lcd.print(good_light ? "P" : "F");

  lcd.setCursor(0,1);
  lcd.print("humid->");
  lcd.print(good_soil ? "P" : "F");

  Serial.println("The plant mood is " + String(good_temp & good_light & good_soil));

  lcd.print(good_temp & good_light & good_soil ? " Pass!" : " !Help!"); 
}

struct sensors {
  float air_humidity;
  float temperature;
  float lightning_intensity;
  float soil_humidity;
};

void loop() {
  if (measure_flag) {
    measure_flag = false; // Reset the flag

    sensors Sensors = { 0, 0, 0, 0 };
    Sensors.air_humidity = dht.readHumidity();
    // Read air temperature (celsius) and humidity (%)
    Sensors.temperature = dht.readTemperature();
  
    // Check for light presence
    Sensors.lightning_intensity = digitalRead(LIGHTPIN);
  
    // Read soil humidity considering reading error
    float sum = 0;
    for (int i = 0; i < 500; ++i) {
      sum += analogRead(SOILHUMIDITYPIN);
    }
    Sensors.soil_humidity = sum / 500;

    good_temp = Sensors.temperature > 18.0 && Sensors.temperature < 28.0;

    // Set the light presence state
    int current_time = millis();
    good_light = Sensors.lightning_intensity == HIGH || dark_start_time + NO_LIGHT_PERIOD < current_time;
    if (!good_light) {
      dark_start_time = millis();
    }

    // Set the soil humidity state
    // If the soil humidity is bigger than 150 the plant needs water
    good_soil = Sensors.soil_humidity < 150;

    // Print on LCD
    print_details_on_lcd(good_temp, good_light, good_soil);

    // Turn on/off the led based on plant state
    plant_mood = good_temp & good_light & good_soil;
    digitalWrite(LEDPIN, plant_mood ? LOW : HIGH);
  }
}

// Timer interrupt service routine
void timerIsr() {
  measure_flag = true;
}

// External interrupt service routine for light sensor
void lightChangeIsr() {
  if (digitalRead(LIGHTPIN) == LOW) {
    dark_start_time = millis(); // Reset the dark start time
  }
}
